/* -*-mode:C;tab-width:8-*-
 * powerd   Monitor the UPS status.
 *
 * Copyright (C) 1993        Miquel van Smoorenburg
 * Copyright (C) 1995        Alessandro Rubini
 * Copyright (C) 2001        Umberto Zappi
 *
 * This program is designed to work with SysVInit-2.4 for Linux and later, by
 * Miquel van Smoorenburg. It may work with other operating systems.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define POWERD_RELEASE "2.1.0"

/* Include header */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <syslog.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/time.h>
#include <time.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <utmp.h>
#include <mntent.h>
#include "powerd.h"


/*---- Globals ------------------------------------------------------*/
int pollingmode=FALSE;
char msgstring[80];
char prgname[64];


/*-------------------------------------------------------------------*/
/* This is the option database */

Option options[]= {

  {"disabled",     0, 0, OT_BOOL,            0, NULL},
  {"debug",        0, 0, OT_BOOL,            0, NULL},  /* for compatibility */
  {"debuglevel",   0, 0, OT_INTEGER,         0, NULL},  /* new (rel. 2.1.0) */
  {"delay",        0, 0, OT_INTEGER,        10, NULL},
  {"sleepnow",     0, 0, OT_INTEGER,         0, NULL},
  {"cycletime",    0, 0, OT_INTEGER,      3600, NULL},  /* unused */
  {"remoteserver", 0, 0, OT_STRING,          0, NULL},
  {"serverport",   0, 0, OT_INTEGER,         0, NULL},
  {"serialline",   0, 0, OT_FILE,            0, NULL},
  {"monitor",      0, 0, OT_RSNAME,   TIOCM_CD, "DCD"},
  {"monitorlow",   0, 0, OT_RSNAME,  TIOCM_CTS, "CTS"}, /* new (rel. 2.1.0) */
  {"failwhen",     0, 0, OT_STR_OR_LEV, 0xffff, "FAIL"},
  {"faillowwhen",  0, 0, OT_STR_OR_LEV, 0xffff, "LOW"}, /* new (rel. 2.1.0) */
  {"okwhen",       0, 0, OT_STRING,          0, "OK"},
  {"shutdown",     0, 0, OT_BOOL,            0, NULL},  /* new (rel. 2.1.0) */
  {"sleepassert",  1, 0, OT_INTEGER,         1, NULL},  /* new (rel. 2.1.0) */
  {"assertlow",    1, 0, OT_RSNAME,          0, NULL},
  {"asserthigh",   1, 0, OT_RSNAME,          0, NULL},
  {"outdtr",       0, 0, OT_LEVEL,       FALSE, NULL},  /* new (rel. 2.1.0) */
  {"outrts",       0, 0, OT_LEVEL,       FALSE, NULL},  /* new (rel. 2.1.0) */
  {"readfifo",     0, 0, OT_FILE,            0, NULL},
  {"spawnmonitor", 0, 0, OT_STRING,          0, NULL},
  {"pollingstep",  0, 0, OT_INTEGER,        20, NULL},
  {"trials",       0, 0, OT_INTEGER,        18, NULL},
  {"timeout",      0, 0, OT_INTEGER,        10, NULL}, 
  {"logfile",      0, 0, OT_STRING,          0, NULL},  /* stderr by default */
#ifdef linux
  {"powerfile",    0, 0, OT_STRING,          0, P_PWRSTAT}, /* new (rel. 2.1.0) */
#endif
  {"",             1, 0, OT_END,             0, NULL}
};

enum optnames {O_DISABLED,O_DEBUG,O_DEBUGLEVEL,O_DELAY,O_SLEEPNOW,O_CYCLETIME,
		 O_REMOTESERVER,O_SERVERPORT,O_SERIALLINE,O_MONITOR,
		 O_MONITORLOW,O_FAILWHEN,O_FAILLOWWHEN,O_OKWHEN,O_SHUTDOWN,
		 O_SLEEPASSERT,O_ASSERTLOW,O_ASSERTHIGH,
		 O_OUTDTR,O_OUTRTS,
		 O_READFIFO,
		 O_SPAWNMONITOR,O_POLLINGSTEP,O_TRIALS,O_TIMEOUT,O_LOGFILE,
#ifdef linux
		 O_POWERFILE,
#endif
               };

/*---- End globals --------------------------------------------------*/


/*===================================================================*
 *     MAIN PROGRAM                                                  *
 *===================================================================*/

int main(int argc, char **argv)
{
  int current_status;
  int sensefd, serverfd=-1;
  int (*handler)(int mode, int arg)=NULL;
  char *prg;
  char runlevel;
  struct assert *pAssertions=NULL; 

  /* Open logging */
  openlog(NULL, LOG_CONS, LOG_DAEMON);

  /* Set program name and pid */
  prg = (strrchr(argv[0], '/') == NULL) ? argv[0] : strrchr(argv[0], '/')+1;
  sprintf(prgname,"%s[%i]", prg, getpid());

  /* Load the options either from configuration file and command line */
  LoadOptions(argc, argv, &pAssertions);
  if (DEBUG_LVL2)
    {
    PrintOptions();
    PrintAssertions(pAssertions);
    }
  
  /* Ok, now we consider running */
  sprintf(msgstring, "Powerd version %s", POWERD_RELEASE);
  if (options[O_DISABLED].iv)
    {
    strcat(msgstring, " (disabled)");
    fprintf(stderr, "%s\n", msgstring);
    }
  MESSAGE(msgstring);

  /* How runlevel are we? */
  runlevel=GetRunlevel();
  sprintf(msgstring,"We are at runlevel %c", runlevel);
  MESSAGE(msgstring);

  /* Check for unique task */
  if ((handler=CheckUniqueTask(options)) == NULL)
    exit(1);

  /* Set pollingmode */
  /* If we check the serial line, we need some polling */
  if (options[O_POLLINGSTEP].iv == 0 && options[O_SERIALLINE].sv != NULL)
    options[O_POLLINGSTEP].iv = 20;
  pollingmode = (options[O_POLLINGSTEP].iv == 0) ? FALSE : TRUE;
    
  /* Redirect messages, if such requested */
  if (options[O_LOGFILE].sv)
    RedirectStderr(options[O_LOGFILE].sv);
    

  /* Daemonize  }:)  */
  if (options[O_LOGFILE].sv != NULL || !DEBUG_LVL1)
    {
    switch(fork()) 
      {
      case 0: /* Child */
	chdir("/");
	/* If we are in runlevel 0, isn't advisable execute setsid(),
	 * otherwise powerd may be killed by shutdown process */
	if (runlevel != '0')
	  setsid();
        sprintf(prgname,"%s[%i]", prg, getpid());
	MESSAGE("Start daemon");
	break;

      case -1: /* Error */
	fprintf(stderr,"%s: %s\n",prgname,strerror(errno));
	exit(1);

      default: /* Parent */
	exit(0);
      }
    }

  /* after daemonize, sleep */
  if (options[O_SLEEPNOW].iv)
    {
    sprintf(msgstring, "Sleeping %d second(s)", options[O_SLEEPNOW].iv);
    MESSAGE(msgstring);
    sleep(options[O_SLEEPNOW].iv);
    }

  /* Default init value */
  current_status = ACT_OK;

  /* Actions based on current runlevel */
  switch(runlevel)
    {
    case 'S':
    case '1':
    case '?':
      /* Maintenance mode: powerd shouldn't be invoked. Exit soon. */
      MESSAGE("We are in maintenance mode. Exit program...(bye!)");
      exit(1);
      break;

    case '6':
      /* System reboot: powerd shouldn't be invoked. Exit soon. */
      MESSAGE("We are rebooting the system. Exit program...(bye!)");
      exit(1);
      break;

    case '0':
      /* System halt: execute UPS's shutdown if required */
      if (options[O_SHUTDOWN].iv && options[O_SERIALLINE].sv)
        {
        /* shutdown UPS */
        MESSAGE("Open serial connection and send a pulse for shutdown our UPS");
        sensefd=CheckSerial(MODE_OPEN,0);
        if (sensefd >= 0)
	  {
          SendPulse(sensefd,pAssertions);
       	  close(sensefd);
	  FreeAssertList(pAssertions);
	  pAssertions=NULL;
	  }
        }
      MESSAGE("Exit program...(bye!)");
      exit(1);
      break;

    default:
      /* No action. Continue ... */
      break;
    }

  
  sensefd=(*handler)(MODE_OPEN,0);
  if (sensefd==ACT_ABORT)
    {
    closelog();
    exit(1);
    }

  fcntl(sensefd,F_SETFL,fcntl(sensefd,F_GETFL)|O_NONBLOCK);

  /* Now, we should open the server node, if needed */
  if (options[O_REMOTESERVER].sv == NULL && options[O_SERVERPORT].iv != 0)
    serverfd = OpenUdpServer(options[O_SERVERPORT].iv);

  while (1)
    current_status=MainCheck(sensefd, serverfd, current_status, 
	                     handler,  options);

  MESSAGE("Aaargh -- fallen out of the loop");
  exit(1); /* shouldn't happen */
}

/*===================================================================*
 *     FUNCTIONS DEFINITION                                          *
 *===================================================================*/

/*-------------------------------------------------------------------*/
/* Deal with options written on the configuration file or passed by
 * the command line.
 * Options specified by the command line have higher priority on the
 * same options specified in the configuration file.
 */

static int LoadOptions(int argc, char **argv, struct assert **p2Assertions)
{
  char *cfg_file=P_CFG_FILE;
  struct assert *pAssertionsCmdLine=NULL; 
  struct stat buf;
  int type;
  FILE *fd_cfg_file=NULL;
  char buffer[256];
  int errors=0;


/*
 * If the last arguments are "-port <port>", the program should accept
 * client connections on that udp port and speak to them.
 */

  if (argc > 2 && strcmp(argv[argc-2],"-port") == 0)
    {
    if (ParseLine("serverport",argv[argc-1],p2Assertions))
      exit(1);
    argc-=2;
    }
  else if (argc > 3 && strcmp(argv[argc-3],"-host") == 0)
    {
    if (ParseLine("remoteserver",argv[argc-2],p2Assertions) ||
	ParseLine("serverport",  argv[argc-1],p2Assertions))
      exit(1);
    argc-=3;
    cfg_file=NULL;
    }


/*
 * If there is a single argument, it is a file. It can be:
 *     a fifo                then, wait for "OK" and/or "FAIL"
 *     a char device         then, check DTR, like to old powerd used to do
 *     a regular file        use it as a config file
 * otherwise issue an error
 */

  if (argc > 1 && strchr(argv[1],'=') == NULL) /* not an option */
    {
    if (lstat(argv[1],&buf) < 0)
      {
      fprintf(stderr,"%s: %s: %s\n",prgname,argv[1],strerror(errno));
      exit(1);
      }
    switch(type = buf.st_mode&S_IFMT)
      {
      case S_IFREG:
	     cfg_file=argv[1]; /* use it */
	     break;

      case S_IFIFO:
	     if (ParseLine("readfifo",argv[1],p2Assertions))
	       exit(1);
	     break;

      case S_IFSOCK:
	     fprintf(stderr,"%s: %s: Is a socket\n",prgname,argv[1]);
	     exit(1);

      case S_IFBLK:
	     fprintf(stderr,"%s: %s: Is a block device\n",prgname,argv[1]);
	     exit(1);

      case S_IFDIR:
	     fprintf(stderr,"%s: %s: %s\n",prgname,argv[1],strerror(EISDIR));
	     exit(1);

      case S_IFCHR: 
	     /*
	      * Ok, it is chr dev, so we must be compatible with powerd-1.1
	      */
	     if (ParseLine("serialline",  argv[1], p2Assertions)
	      || ParseLine("monitor",     "DCD",   p2Assertions)
	      || ParseLine("failwhen",    "low",   p2Assertions)
	      || ParseLine("pollingstep", "20",    p2Assertions))
	       exit(1);
	     break;

      default:
	     fprintf(stderr,"%s: %s: Can't manage such node\n",prgname,argv[1]);
	     exit(1);
      }
    argc--; argv++;
    }

/*
 * Read the configuration file.
 */

  if (cfg_file)
    {
    fd_cfg_file=fopen(cfg_file,"r");
    if (!fd_cfg_file)
      {
      fprintf(stderr,"%s: %s: %s\n",prgname,cfg_file,strerror(errno));
      exit(1);
      }
    else
      {
      while (fgets(buffer,sizeof(buffer),fd_cfg_file))
        errors=ParseLine(buffer,NULL,p2Assertions);
      fclose(fd_cfg_file);
      if (errors)
        exit(errors);
      }
    }

/*
 * Get options from args
 */

/* 
 * The options from command line have precedence over the options on the
 * parameter file
 */
  if (DEBUG_LVL1)
    fprintf(stderr, "------ options from command line ------\n");

  CleanUsage(options);
  while (argc > 1)
    {
    errors=ParseLine(argv[1], NULL, &pAssertionsCmdLine);
    argc--; argv++;
    }
  if (errors)
    exit(errors);
  
  if (pAssertionsCmdLine != NULL)    /* assertions presents on the cmd line */
    {
    FreeAssertList(*p2Assertions);   /* free the original assertion list */
    *p2Assertions=pAssertionsCmdLine;/* move start pointer of new assertion l.*/
    pAssertionsCmdLine=NULL;
    }
  
  return 0;
}

/*-------------------------------------------------------------------*/
/* This parses a single argument, either form the command line or
 * from a line in the configuration file. Escape sequences are not
 * interpreted, becuase it tries to remain small. The return value
 * is the number of errors up to now. Arguments are the oprion name and,
 * optional, the suggested value. If it is null, then the nameis split
 * at an equal sign or at the first blank space.
 */

static int ParseLine(char *opt,char *svalue,struct assert **p2Assertion)
{
  static int parserrors=0;
  char *value,*ptr;
  Option *optr;
  struct stat buf;
  struct assert *pCurrAssertion;

  if (opt[strlen(opt)-1]=='\n') opt[strlen(opt)-1]='\0';
  if (svalue && svalue[strlen(svalue)-1]=='\n') svalue[strlen(svalue)-1]='\0';

  PDEBUG_L9((stderr,"\tParseLine: \"%s\" \"%s\"\n",opt,svalue));

  /* retrieve the value */
  if (svalue)        /* preparsed */
    value=svalue; 
  else               /* parse it */
    {
    while (isspace(*opt))
      opt++; 
    if (*opt=='#' || *opt=='\0')
      return parserrors;
    if (  !(ptr=strchr(opt,'='))    /* '=' separates option and value */
       && !(ptr=strchr(opt,' '))    /* or it is a blank */
       && !(ptr=strchr(opt,'\t')))  /* or a tab */
      {ptr=""; value="";}           /* or there's no value */
    else
      {*ptr='\0';value=ptr+1;}

    while (isspace(*value)) /* strip any blank */
      value++;
    for (ptr=opt+strlen(opt)-1; isspace(*opt); *(opt--)='\0')
      ;
    }

  /* Now, if the first *and* the last are quotes, remove them */
  if (*value=='"' && value[strlen(value)-1]=='"')
     {value++; value[strlen(value)-1]='\0';}

  /* MISSING: special chars (escape sequences) */
  if (DEBUG_LVL1)
    fprintf(stderr,"%20s --> %s ",opt,value);

/*---------------------------- Now do the job */

  for (optr=options; optr->type!=OT_END && strcmp(optr->name,opt); optr++)
    /* nothing */ ;

  if (optr->usage++ > 0 && !(optr->multi_allowed))
    {
    fprintf(stderr,"%s: \"%s\" is multiply defined\n",prgname,opt);
    return ++parserrors;
    }
     
  switch(optr->type)
    {
    case OT_BOOL:
      if (value == NULL || value[0] == '\0' || tolower(value[0]) != 'n')
	optr->iv=TRUE;
      else
	optr->iv=FALSE;
      if (optr == options+O_DEBUG)  /* debug and debuglevel live together */
	options[O_DEBUGLEVEL].iv = (optr->iv) ? 1 : 0;
      break;

    case OT_STRING:
      optr->sv=strdup(value);
      if (optr == options+O_REMOTESERVER)
	ParseLine("delay","0",p2Assertion);
      break;
      
    case OT_FILE:
      optr->sv=strdup(value);
      if (lstat(value,&buf))
	{
	fprintf(stderr,"%s: %s: %s\n",prgname,value,strerror(errno));
	return ++parserrors;
	}
      optr->iv = buf.st_mode&S_IFMT;
      break;

    case OT_RSNAME:
      optr->sv=strdup(value);
      optr->iv='!';  /* just a warning */
      if (!strcasecmp(value,"rts")) optr->iv = TIOCM_RTS | BIT_OUT;
      if (!strcasecmp(value,"dtr")) optr->iv = TIOCM_DTR | BIT_OUT;
      if (!strcasecmp(value,"dcd")) optr->iv = TIOCM_CD;
      if (!strcasecmp(value,"dsr")) optr->iv = TIOCM_DSR;
      if (!strcasecmp(value,"cts")) optr->iv = TIOCM_CTS;

      /* if a pulse is required, we must record it, to execute after opening */
      if (optr == options+O_ASSERTHIGH || optr == options+O_ASSERTLOW)
	{
	if (!(optr->iv&BIT_OUT))
	  {
	  fprintf(stderr,"%s: can't assert \"%s\"\n",prgname,optr->sv);
          return ++parserrors;
	  }
	optr->iv &= ~BIT_OUT;
	pCurrAssertion=AppendAssertList(p2Assertion);
	if (pCurrAssertion == NULL) 
	  {
	  fprintf(stderr,"%s: malloc(): %s\n",prgname,strerror(errno));
	  return ++parserrors;
	  }
	pCurrAssertion->bit=optr->iv;
	pCurrAssertion->val= (optr == options+O_ASSERTHIGH ? 0xffff : 0);
	pCurrAssertion->sleep=options[O_SLEEPASSERT].iv;
	pCurrAssertion->next=NULL; 
	}
      break;

    case OT_STR_OR_LEV:
      optr->sv=strdup(value);
      optr->iv='!';  /* just a warning */
      if (!strcasecmp(value,"low"))  optr->iv=0;
      if (!strcasecmp(value,"high")) optr->iv=0xffff;
      break;

    case OT_LEVEL:
      optr->sv=strdup(value);
      if (strcasecmp(value,"low") == 0) 
	optr->iv=FALSE;
      else if (strcasecmp(value,"high") == 0) 
	optr->iv=TRUE;
      else
        {
        fprintf(stderr,"%s: \"%s\" level \"%s\" unknown\n",prgname,opt,value);
        return ++parserrors;
        }
      break;
    
    case OT_INTEGER:
      optr->iv=atoi(value);
      if (optr == options+O_DEBUGLEVEL) /* debug and debuglevel live together */
	options[O_DEBUG].iv = (optr->iv != 0) ? TRUE : FALSE;
#ifdef EXPERIMENTAL
      else if (optr == options+O_SLEEPASSERT)
        {
        /* Modify sleep time in the last element of assertion list */
	if ((pCurrAssertion=GetLastAssert(*p2Assertion)) != NULL)
	  pCurrAssertion->sleep = optr->iv;
	}
#endif      
      break;
      
    default:
      fprintf(stderr,"%s: Unknown option \"%s\"\n",prgname,opt);
      ++parserrors;
      break;
    }
  if (DEBUG_LVL1)
    fprintf(stderr," (%i)\n",optr->iv);

  return parserrors;
}

/*-------------------------------------------------------------------**/
/* Print all options 
 * This code is compiled only if DEBUG is defined
 */

static void PrintOptions(void)
{
#ifdef DEBUG
  Option *pOpt;

  fprintf(stderr, "\n\nPrint all options\n\n");
  fprintf(stderr, "%20s %20s %8s\n", "Name", "Value (string)", "Val.%d");
  fprintf(stderr, "--------------------------------------------------\n");

  for (pOpt=options; pOpt->type != OT_END; pOpt++)
    fprintf(stderr, "%20s %20s (%6d)\n", pOpt->name, pOpt->sv, pOpt->iv);

  fprintf(stderr, "--------------------------------------------------\n");

#else
  fprintf(stderr, "For this option you must compile with -DDEBUG\n");
#endif
}

/*-------------------------------------------------------------------**/
/* Print the assertion's list. 
 * This code is compiled only if DEBUG is defined
 */

static void PrintAssertions(struct assert *pAssert)
{
#ifdef DEBUG
  fprintf(stderr, "\n\nPrint the assertions\'s list\n\n");
  fprintf(stderr, "%10s %6s %6s %6s\n", "SignalName", "Bit", "Value", "Sleep");
  fprintf(stderr, "-------------------------------\n");

  for ( ; pAssert != NULL; pAssert=pAssert->next)
    {
    switch(pAssert->bit)
      {
      case TIOCM_DTR:
        fprintf(stderr, "%10s ", "DTR");
	break;
      case TIOCM_RTS:
        fprintf(stderr, "%10s ", "RTS");
	break;
      default:
	fprintf(stderr, "%10s ", "Unknown");
	break;
      }
    fprintf(stderr, "%#06x %#06x %6d\n", 
	pAssert->bit, pAssert->val, pAssert->sleep);
    }

  fprintf(stderr, "-------------------------------\n");

#else
  fprintf(stderr, "For this option you must compile with -DDEBUG\n");
#endif
}

/*-------------------------------------------------------------------*/
/* Check for unique task.
 * Return a pointer to function if ok, NULL elsewise
 */

static int (*CheckUniqueTask(Option options[]))(int,int)
{
  int i;
  int (*handler)(int mode, int arg)=NULL;


  handler=NULL;
  i=0;
  if (options[O_REMOTESERVER].sv) 
    {
    handler=CheckSocket;
    i++;
    if(DEBUG_LVL1)
      fprintf(stderr,"%s: should reach %s\n",
	      prgname, options[O_REMOTESERVER].sv);
    }
  if (options[O_READFIFO].sv)
    {
    handler=CheckFIFO;
    i++;
    if(DEBUG_LVL1)
      fprintf(stderr,"%s: should read fifo %s\n",
	      prgname, options[O_READFIFO].sv);
    }
  if (options[O_SERIALLINE].sv)
    {
    handler=CheckSerial;
    i++;
    if(DEBUG_LVL1)
      fprintf(stderr,"%s: should rd serial %s\n",
	      prgname,options[O_SERIALLINE].sv);
    }
  if (options[O_SPAWNMONITOR].sv)
    {
    handler=CheckChild;
    i++;
    if(DEBUG_LVL1)
      fprintf(stderr,"%s: should rd child %s\n",
	      prgname,options[O_SPAWNMONITOR].sv);
    }

  if (i == 0) fprintf(stderr,"%s: Nothing to check\n",prgname);
  if (i >  1) fprintf(stderr,"%s: Ambiguous setup\n",prgname);

  return (i==1) ? handler : NULL;
}

/*-------------------------------------------------------------------*/
/* Change the file associated with a standard error with logfile.
 * I other words redirect stderr on logfile.
 */

static void RedirectStderr(const char *logfile)
{
  int fd;

  if (DEBUG_LVL1) 
    {
    sprintf(msgstring,"Trying to redirect stderr to \"%s\"", logfile);
    MESSAGE(msgstring);
    }
  if ((fd=open(logfile, O_WRONLY|O_APPEND)) < 0)
    MESSAGE("Failed to redirect stderr");
  else
    {
    close(fd);
    if (!freopen(logfile,"a",stderr))
      printf("%s: %s: %s\n", prgname, logfile, strerror(errno));
    else
      setvbuf(stderr, NULL, _IONBF, 0); /* prevent delays */
    }
}

/*-------------------------------------------------------------------*/
/* Open an udp connection as server and bind the opened socket.
 * Return a valid file handler or -1 in case of error.
 */

static int OpenUdpServer(int udp_port)
{
  int fd_udp;
  int len;
  struct sockaddr_in serveraddr;

  if (DEBUG_LVL1)
    fprintf(stderr,"Opening udp socket...\n");

  if ((fd_udp=socket(AF_INET, SOCK_DGRAM, 0)) == -1)
    PERROR("socket()");
  else
    {
    serveraddr.sin_family = AF_INET;
    serveraddr.sin_addr.s_addr = INADDR_ANY;
    serveraddr.sin_port = htons(udp_port);
    if (bind(fd_udp, (struct sockaddr *)&serveraddr, sizeof(serveraddr)) < 0)
      {
      PERROR("bind()");
      close(fd_udp);
      fd_udp = -1;
      }
    else
      {
      len = sizeof(serveraddr);
      if (getsockname(fd_udp, (struct sockaddr *)&serveraddr, &len))
        {
        PERROR("getsockname()");
        close(fd_udp);
        fd_udp = -1;
        }
      else if (DEBUG_LVL1) 
        fprintf(stderr, "Got udp port %i\n", ntohs(serveraddr.sin_port));
      }
    }

  return fd_udp;
}

/*-------------------------------------------------------------------*/
/* Execute the main check of discovery every change on power status
 * and if necessary call the PowerFail() function.
 * Return the status discovery.
 */

static int MainCheck(int sensefd, int serverfd, int current_status,
                      int (*handler)(int,int), Option options[])
{
  int pending, newpending;
  int status, newstatus;
  int maxfd;
  fd_set selSet, newSet;
  static struct client *clientlist=NULL;
  struct timeval timeout;
  struct timeval to;
	  

  FD_ZERO(&selSet);
  FD_SET(sensefd,&selSet);
  if (serverfd>=0)
    FD_SET(serverfd,&selSet);
  /* select() may update the timeout parameter */
  timeout.tv_sec=options[O_POLLINGSTEP].iv;
  timeout.tv_usec=0;
  maxfd = (sensefd>serverfd) ? sensefd : serverfd;
  pending=select(maxfd+1,&selSet,NULL,NULL,&timeout);
  if (pollingmode && !FD_ISSET(sensefd,&selSet)) 
    {
    pending++;
    FD_SET(sensefd,&selSet);
    }

  if (pending < 1)
    ;
  else if (FD_ISSET(sensefd,&selSet))
    {
    pending--;
    status=(*handler)(MODE_READ,sensefd);
    if (status==ACT_NOTHING)
      ;
    else if (status==current_status)
      PDEBUG_L9((stderr,"%s: Nothing new\n",CurrTime()));
    else
      {
      /*
       * Try to sleep for a while, and if there's sth new, read it.
       * I've had a lot of problem in avoiding race conditions, but
       * it should work well now.
       */
      if (status==ACT_FAILLOW)
	MESSAGE("Advisory: the power is failing (and low battery)");
      else if (status==ACT_FAIL)
	MESSAGE("Advisory: the power is failing");
      else
	MESSAGE("Advisory: the power is back");

      newpending=0;
      newstatus=status; /* default */
      if (options[O_DELAY].iv)
	{
        if (DEBUG_LVL1)
	  MESSAGE("Delaying action...");
        FD_ZERO(&newSet);
       	FD_SET(sensefd,&newSet);
        to.tv_sec=options[O_DELAY].iv;
        to.tv_usec=0;
	newpending=select(sensefd+1, &newSet, NULL, NULL, &to);
	}
      if(newpending == 1 || pollingmode)
	newstatus=(*handler)(MODE_READ, sensefd); /* if it changes */
      if (newstatus == ACT_NOTHING)
	{
	newstatus=status;
	if (DEBUG_LVL1)
	  MESSAGE("Can't get delayed status. It's forced to info status");
	}
      if (DEBUG_LVL1)
	{
	sprintf(msgstring,"current \'%c\', info \'%c\', delayed \'%c\'",
 	  current_status,status,newstatus);
	MESSAGE(msgstring);
	}
      if (newstatus == current_status)
	MESSAGE("'Twas a glitch");
      else
	{
	if (TellClients(newstatus, serverfd, clientlist) == 0)
	  {
	  FreeClientList(clientlist);
	  clientlist=NULL;
	  }
	PowerFail(newstatus);
	current_status=newstatus;
	}
      } 
    } 

  if (pending > 0 && serverfd >= 0 && FD_ISSET(serverfd,&selSet))
    {
    pending--;
    ManageClient(serverfd, &clientlist, current_status);
    }

  return current_status;
}

/*-------------------------------------------------------------------*/
/* Return the current runlevel read from the utmp system file
 */

static unsigned char GetRunlevel(void)
{
  unsigned char retval;
  struct utmp ut;
  struct utmp *put;

  setutent();	/* rewind the file pointer to the beginning of the utmp file */
  ut.ut_type = RUN_LVL;
  if ((put=getutid(&ut)) == NULL) 
    retval = '?';		/* unknown runlevel */
  else
    retval = put->ut_pid%256;	/* ut_pid = prev_runlev*256 + curr_runlev */
  
  endutent();	/* close the utmp file */
  return retval;
}

/*-------------------------------------------------------------------*/
/* Return the device of pathname passed in argument
 */

static dev_t DeviceOf(const char *pathname)
{
  dev_t retval=-1;
  struct stat buf;
  char mypath[MAX_SZPATH];
  char *pch;
  int try=0;
  int exitloop=FALSE;
  
  strcpy(mypath, pathname);
  while (exitloop == FALSE)
    {
    if (stat(mypath, &buf) == 0)
      {
      retval = buf.st_dev;	/* success */
      exitloop = TRUE;
      }
    else if ((pch=strrchr(mypath, '/')) == NULL) /* search for the last '/' */
      {
      /* no '/' found ... */
      if (try == 0)
        {
        /* ... but we'll try stat() with current directory */
        strcpy(mypath, ".");
        try = 1;
        }
      else
	{
        retval = -1;	/* stat(".", &buf) not successfull ?!?! */
	exitloop = TRUE;
	}
      }
    else if (pch == mypath)
      {
      /* we are arrived at root filesystem "/" ... */
      if (try == 0)
        {
        /* ... and we'll try stat() */
        *(pch+1) = '\0';
        try = 1;
        }
      else
	{
        retval = -1;	/* stat("/", &buf) not successfull ?!?! */
	exitloop = TRUE;
	}
      }
    else
      *pch = '\0';	/* delete path from last '/' */
    }

  return retval;
}

/*-------------------------------------------------------------------*/
/*
 * Return the pointer to a struct mntent with reference to pathname.
 * This struct is allocated in a static area of memory and is
 * overwritten by subsequent calls to getmntent()
 */

static struct mntent *MntentOf(const char *pathname)
{
  struct mntent *pmntent=NULL;
  struct stat buf;
  dev_t device;
  FILE *fdmntent=NULL; 

  device=DeviceOf(pathname);
  fdmntent=setmntent(P_MNT_INFO, "r");	/* open for read on P_MNT_INFO */
  if (fdmntent == NULL)
    {
    PERROR("setmntent()");
    exit (1);
    }
  else
    {
    do
      {
      if ((pmntent=getmntent(fdmntent)) == NULL)
        {
        MESSAGE("getmntent() return NULL");
        exit(1);
        }
      else if (stat(pmntent->mnt_dir, &buf) < 0)
        {
        PERROR("stat()");
        exit(1);
        }
      }
    while(buf.st_dev != device);
    endmntent(fdmntent);	/* close read on P_MNT_INFO */
    }
  
  return pmntent;
}

/*-------------------------------------------------------------------*/
/*
 * Return mount point of the filesystem with reference to pathname
 */

static char *MountPointOf(const char *pathname)
{
  return (MntentOf(pathname))->mnt_dir;
}

/*-------------------------------------------------------------------*/
/* Return true if the current status of the filesystem with reference
 * to pathname is mounted with read-only option
 */

static int IsROFS(const char *pathname)
{
  return (hasmntopt(MntentOf(pathname), MNTOPT_RO) != NULL) ? TRUE : FALSE;
}

/*-------------------------------------------------------------------*/
/* Return the current status of the filesystem (read-only or read-write)
 * with reference to pathname
 */

static int GetMountFS(const char *pathname)
{
  return (IsROFS(pathname)) ? FS_RO : FS_RW;
}

/*-------------------------------------------------------------------*/
/* Format the current time, so reports on stderr are more meaningful
 * when retrieved by a redirection file
 */
static char *CurrTime(void)
{
  static time_t mytime;
  static char str_time[32];

  time(&mytime);
  strftime(str_time,sizeof(str_time),"%a %b %d %H:%M:%S",localtime(&mytime));
  return str_time;
}

/*-------------------------------------------------------------------*/
/* Clean field usage of the struct Option
 */

static void CleanUsage(Option options[])
{
  while (options->type != OT_END)
    (options++)->usage = 0;
}

/*-------------------------------------------------------------------*/
/* Alloc memory for struct assert at the end of list.
 * Return the pointer of the allocated area or NULL if memory exausted
 */

static struct assert *AppendAssertList(struct assert **p2Assert)
{
  struct assert *retval;
  
  if (*p2Assert != NULL)
    retval=AppendAssertList(&(*p2Assert)->next);
  else
    retval=*p2Assert=malloc(sizeof(struct assert));
  
  return retval;    
}

#ifdef EXPERIMENTAL
/*-------------------------------------------------------------------*/
/* Retrieve the last element of assertions list.
 * Return the pointer at last element or NULL if list is void.
 */

static struct assert *GetLastAssert(struct assert *pAssert)
{
  struct assert *retval;

  if (pAssert == NULL)
    /* No elements */
    retval = NULL;
  else if (pAssert->next == NULL)
    /* We are at last element */
    retval = pAssert;
  else
    retval = GetLastAssert(pAssert->next);

  return retval;
}
#endif

/*-------------------------------------------------------------------*/
/* Free memory from assertions list
 */

static void FreeAssertList(struct assert *pAssert)
{
  if (pAssert != NULL)
    {
    FreeAssertList(pAssert->next);
    free(pAssert);
    }
}


/*===================================================================*
 * This section embeds the engines, the different monitoring         *
 * utilities.                                                        *
 * You're free to add new engines.                                   *
 *===================================================================*/

/*-------------------------------------------------------------------*/
/* Check if the string argument contain information on power status
 * and return what is discovery.
 */

static int CheckString(char *buffer)
{
  if (strstr(buffer,options[O_FAILWHEN].sv)) return ACT_FAIL;
  if (strstr(buffer,options[O_FAILLOWWHEN].sv)) return ACT_FAILLOW;
  if (strstr(buffer,options[O_OKWHEN].sv)) return ACT_OK;
  return ACT_NOTHING;      
}

/*-------------------------------------------------------------------*/
/* Open or read the serial stream.
 * If mode == MODE_OPEN, willl open the serial stream and is returned
 * or the file handler or a negative number (if fail happened).
 * If mode == MODE_READ, will read the stream and is returned
 * an action value (power status).
 */

static int CheckSerial(int mode, int arg)
{
int fd, bit;
int retval;
static int bitmask, prevval=BIT_OUT; /* impossible value */
static int bitfail, valfail;
static int bitfaillow, valfaillow;

  switch(mode)
    {
    case MODE_OPEN:
      /* Open the serial stream */
      if (options[O_SERIALLINE].iv != S_IFCHR)
        {
        sprintf(msgstring,"%s is not a char device",options[O_SERIALLINE].sv);
        MESSAGE(msgstring);
        retval = ACT_ABORT;
        }
      else if (options[O_MONITOR].iv=='!' || options[O_FAILWHEN].iv=='!')
        {
        sprintf(msgstring,"Wrong options for \"monitor\" or \"failwhen\"");
        MESSAGE(msgstring);
        retval = ACT_ABORT;
        }
      else if (options[O_MONITOR].iv & BIT_OUT)
        {
        sprintf(msgstring,"Can't monitor an outgoing line (\"%s\")",
  	        options[O_MONITOR].sv);
        MESSAGE(msgstring);
        retval = ACT_ABORT;
        }
      else if (options[O_MONITORLOW].iv=='!' || options[O_FAILLOWWHEN].iv=='!')
        {
        sprintf(msgstring,"Wrong options for \"monitorlow\" or "
	   "\"faillowwhen\"");
        MESSAGE(msgstring);
        retval = ACT_ABORT;
        }
      else if (options[O_MONITORLOW].iv & BIT_OUT)
        {
        sprintf(msgstring,"Can't monitor an outgoing line (\"%s\")",
 	        options[O_MONITORLOW].sv);
        MESSAGE(msgstring);
        retval = ACT_ABORT;
        }
      else if ((fd = open(options[O_SERIALLINE].sv, O_RDWR | O_NDELAY)) < 0) 
        {
        PERROR(options[O_SERIALLINE].sv);
        retval = ACT_ABORT;
        }
      else
	{
        /* set the output bits */
        ioctl(fd,TIOCMGET,&bit);
        if (options[O_OUTDTR].iv == TRUE)
          bit |= TIOCM_DTR;
        else
          bit &= (~TIOCM_DTR);
        if (options[O_OUTRTS].iv == TRUE)
          bit |= TIOCM_RTS;
        else
          bit &= (~TIOCM_RTS);
        ioctl(fd,TIOCMSET,&bit);

        /* opened, now we must remember what to look for */
        bitfail    = options[O_MONITOR].iv;
        bitfaillow = options[O_MONITORLOW].iv;
        bitmask    = bitfail | bitfaillow;
        valfail    = options[O_FAILWHEN].iv ? bitfail : 0;
        valfaillow = options[O_FAILLOWWHEN].iv ? bitfaillow : 0;
        retval = fd;
	}
      break;

    case MODE_READ:
      /* real monitoring */
      if (DEBUG_LVL3)
	MESSAGE("Reading serial line");
      ioctl(arg,TIOCMGET,&bit);
      bit = (bit&bitmask);
      if (bit!=prevval)
        MESSAGE("Line change detected");
      prevval=bit;
      if ((bit&bitfaillow) == valfaillow)
        retval = ACT_FAILLOW;
      else if ((bit&bitfail) == valfail)
        retval = ACT_FAIL;
      else
        retval = ACT_OK;
      break;

    default:
      MESSAGE("CheckSerial(): PROGRAM ERROR");
      exit(1);
      break;
    }

  return retval;
}

/*-------------------------------------------------------------------*/
/* Open or read the socket stream.
 * If mode == MODE_OPEN, willl open the socket stream and is returned
 * or the file handler or a negative number (if fail happened).
 * If mode == MODE_READ, will read the stream and is returned
 * an action value (power status).
 */

static int CheckSocket(int mode, int arg)
{
  int retval=0x7fff;
  int i, j, len;
  int fd;
  int save_pollingmode;
  int server_reply;
  char *host=options[O_REMOTESERVER].sv;
  fd_set set;
  struct timeval to;
  struct powerd_message buf;
  static struct hostent *hostPtr;
  static struct sockaddr_in remote;

  switch(mode)
    {
    case MODE_OPEN:
      /* Open the socket stream */
      sprintf(msgstring,"Connecting to %s",host);
      MESSAGE(msgstring);
      if ((hostPtr=gethostbyname(host)) == NULL)
	{
       	PERROR("gethostbyname()");
       	retval = ACT_ABORT;
       	}
      else
	{
        bzero((char *) &remote, sizeof(remote));
        bcopy(hostPtr->h_addr,(char *)&remote.sin_addr,hostPtr->h_length);
        remote.sin_family=hostPtr->h_addrtype;
        remote.sin_port=(unsigned short)htons(options[O_SERVERPORT].iv);
        if ((fd=socket(AF_INET, SOCK_DGRAM, 0)) < 0)
	  {
       	  PERROR("socket()");
       	  retval = ACT_ABORT;
       	  }
	}
      
      server_reply = FALSE;
      for (i=0; 
	   retval >= 0 && server_reply == FALSE && i < options[O_TRIALS].iv;
	   i++)
        {
        if (i != 0)
	  MESSAGE("still trying...");

        /* On the server must run the same version (or up) of powerd */
        buf.cmd=ACT_ASK;
        strcpy(buf.magic,CURRENT_MAGIC);
        if (sendto(fd, &buf, sizeof(buf), 0,
	      (struct sockaddr *)&remote, sizeof(remote)) < 0)
	  PERROR("sendto()");
	else
	  {
          FD_ZERO(&set);
          FD_SET(fd,&set);
          to.tv_sec=options[O_TIMEOUT].iv;
          to.tv_usec=0;
          if (select(fd+1, &set, NULL, NULL, &to) <= 0)
	    MESSAGE("No reply from server");
	  else
	    {
	    /* Read reply from server */
            PDEBUG_L9((stderr,"%s: Readable\n",CurrTime()));
	    save_pollingmode = pollingmode;	/* Save the pollingmode */
	    pollingmode = FALSE;
            if (CheckSocket(MODE_READ, fd) != ACT_NOTHING)
              server_reply = TRUE;		/* Successfull ! */
	    pollingmode = save_pollingmode;	/* Restore the pollingmode */
	    }
	  }
        }
      if (retval >= 0) 
	{
        if (i == options[O_TRIALS].iv) 
          {
          close(fd);
	  sprintf(msgstring, "FAIL connection to %s !", host);
	  MESSAGE(msgstring);
          retval = ACT_ABORT;
          }
	else
	  {
	  if (i != 0)
       	    MESSAGE("Succeeded");
          retval = fd;
	  }
	}
      break;

    case MODE_READ:
      /* We got a real message (or have to ask about) */
      fd=arg;
      if (pollingmode)
        {
        FD_ZERO(&set);
       	FD_SET(fd,&set);
	to.tv_sec = 0;
	to.tv_usec = 0;
        if (select(fd+1,&set,NULL,NULL,&to) < 1) /* empty, go on asking */
          {
          if(DEBUG_LVL3)
	    MESSAGE("Polling server");
          buf.cmd=ACT_ASK;
	  strcpy(buf.magic,CURRENT_MAGIC); 
          if (sendto(fd, &buf, sizeof(buf), 0,
		(struct sockaddr *)&remote, sizeof(remote)) < 0)
	    PERROR("sendto()");
          retval = ACT_NOTHING;   /* select will bring us the reply */
          }
        }

      if (retval >= 0)
	{
        len=sizeof(remote);
        j=recvfrom(fd,&buf,sizeof(buf), 0, (struct sockaddr *)&remote, &len);
        if (j < 0)
       	  {
	  PERROR("recvfrom()");
       	  retval = ACT_NOTHING;
       	  }
	else
          PDEBUG_L9((stderr,"%s: Got %i bytes (action %i) from port %i on %s\n",
	           CurrTime(), j, buf.cmd, ntohs(remote.sin_port),
	           inet_ntoa(remote.sin_addr)));
	}

      if (retval >= 0)
	{
        /* Check it really is the server. FIXME -- it should obviously be*/
        if (ntohs(remote.sin_port) != options[O_SERVERPORT].iv ||
            memcmp(&remote.sin_addr, hostPtr->h_addr, hostPtr->h_length) != 0)
          {
          MESSAGE("Bogus message from unknown sender");
	  retval = ACT_NOTHING;
	  }
        else if (j != sizeof(buf) || strcmp(buf.magic,CURRENT_MAGIC))
          {
       	  MESSAGE("Bogus message from server");
       	  retval = ACT_NOTHING;
	  }
        else
	  {
	  if (DEBUG_LVL1)
       	    MESSAGE("Got packet");
          retval = buf.cmd;
	  }
	}
      break;

    default:
      MESSAGE("CheckSocket(): PROGRAM ERROR");
      exit(1);
      break;
    }

  return retval;
}

/*-------------------------------------------------------------------*/
/* Open or read the fifo stream.
 * If mode == MODE_OPEN, willl open the fifo stream and is returned
 * or the file handler or a negative number (if fail happened).
 * If mode == MODE_READ, will read the stream and is returned
 * an action value (power status).
 */

static int CheckFIFO(int mode, int arg)
{
  #define RDBUFLEN 256
  int retval;
  char buffer[RDBUFLEN];
  int i;

  switch(mode)
    {
    case MODE_OPEN:
      /* Open the FIFO stream */
      if (options[O_READFIFO].iv != S_IFIFO)
        {
        sprintf(msgstring,"%s is not a fifo",options[O_READFIFO].sv);
        MESSAGE(msgstring);
        retval = ACT_ABORT;
        }
      else
	{
        sprintf(msgstring,"waiting for \"%s\" or \"%s\" or \"%s\"\n",
	        options[O_FAILWHEN].sv, options[O_FAILLOWWHEN].sv,
	        options[O_OKWHEN].sv);
        MESSAGE(msgstring);
        retval = open(options[O_READFIFO].sv,O_RDONLY|O_NONBLOCK);
	}
      break;

    case MODE_READ:
      /* Read from FIFO */
      i=read(arg,buffer,RDBUFLEN-1);
      buffer[i]='\0';
      if (i==0)
        {
	MESSAGE("Connection is lost\n");
       	exit(1);
	}
      PDEBUG_L9((stderr,"%s: Got \"%s\"\n",CurrTime(),buffer));
      if (DEBUG_LVL3)
       	MESSAGE("Got new info");
      retval = CheckString(buffer);
      break;

    default:
      MESSAGE("CheckFIFO(): PROGRAM ERROR");
      exit(1);
      break;
    }

  return retval;
  #undef RDBUFLEN
 }

/*-------------------------------------------------------------------*/
/* Open or read the fifo stream.
 * If mode == MODE_OPEN, willl open the fifo stream and is returned
 * or the file handler or a negative number (if fail happened).
 * If mode == MODE_READ, will read the stream and is returned
 * an action value (power status).
 */

static int CheckChild(int mode, int arg)
{
  int retval;
  FILE *f;

  switch(mode)
    {
    case MODE_OPEN:
      if ((f=popen(options[O_SPAWNMONITOR].sv,"r")) == NULL)
       	{
       	PERROR(options[O_SPAWNMONITOR].sv);
       	retval = ACT_ABORT;
	}
      else
        retval = fileno(f);
      break;

    case MODE_READ:
      retval = CheckFIFO(mode,arg);
      break;

    default:
      MESSAGE("CheckChild(): PROGRAM ERROR");
      exit(1);
      break;
    }

  return retval;
}    

/*-------------------------------------------------------------------*/
/* Send a pulse on serial line
 */

static void SendPulse(int fd, struct assert *pAssert)
{
  struct assert *curr;
  int bit;
  
  /* perform all the assertions we're required to */
  curr=pAssert;
  while (curr != NULL)
    {
    if (DEBUG_LVL1)
      fprintf(stderr,"%s: Assert %i to %i\n", CurrTime(),curr->bit,curr->val);
    ioctl(fd,TIOCMGET,&bit);
    bit &= (~(curr->bit));		/* bit reset */
    bit |= (curr->val & curr->bit);     /* bit set */
    ioctl(fd,TIOCMSET,&bit);
    sleep(curr->sleep);
    curr=curr->next;
    }
}

/*-------------------------------------------------------------------*/
/*
 * Translate action in dependence of magic
 */

static int TranslateAction(int action, char *magic)
{
  int retval;
  
  if (strcmp(magic,CURRENT_MAGIC) == 0)
    retval=action;
  else if (strcmp(magic,MAGIC_2_0) == 0)
    switch(action)
      {
      case ACT_OK:
	retval=ACT_OK_2_0;
	break;

      case ACT_FAIL:
      case ACT_FAILLOW:
	retval=ACT_FAIL_2_0;
	break;

      default:
	retval=action;
	break;
      }
  else
    retval=action;	/* FIXME: is it better return ACT_NOTHING ? */

  return retval;
}

/*-------------------------------------------------------------------*/
/* This handles clients, for udp
 * This is only called on "connection". The client must be
 * registered on the list, and the current power status is returned to it.
 * Return negative number on error.
 */

static int ManageClient(int fd, struct client **list, int current_status)
{
  int retval=0;
  int len;
  int count;
  struct client *clnt=NULL;
  struct sockaddr_in addr;
  struct powerd_message buf;

  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = INADDR_ANY;
  len=sizeof(addr);
  count=recvfrom(fd, &buf, sizeof(buf), 0, (struct sockaddr *)&addr, &len);
  if (count < 0)
    {
    PERROR("recvfrom()");
    retval = -1;
    }
  else if (count != sizeof(buf))
    {
    sprintf(msgstring,"invalid message size (%i) from %s",
	    count, inet_ntoa(addr.sin_addr));
    MESSAGE(msgstring);
    retval = -1;
    }
  else if ((strcmp(buf.magic,CURRENT_MAGIC) == 0 && buf.cmd == ACT_ASK) ||
           (strcmp(buf.magic,MAGIC_2_0) == 0     && buf.cmd == ACT_ASK_2_0))
    ; /* ok */
  else
    {
    MESSAGE("Bad packet received");
    retval = -1;
    }

  if (retval >= 0)
    {
    /* Identify this client in the list (only the client, not the port */
    for (clnt=*list; clnt != NULL; clnt=clnt->next)
      if (memcmp(&(clnt->addr.sin_addr),
	         &(addr.sin_addr), sizeof(addr.sin_addr)) == 0) /* match */
        break;

    if (clnt != NULL)
      {
      /* Client matched */
      if (DEBUG_LVL1)
        {
        sprintf(msgstring,"Client %s is polling",inet_ntoa(addr.sin_addr));
        MESSAGE(msgstring);
        }
      }
    else
      {
      /* Client not found */
      if (DEBUG_LVL1)
        {
        sprintf(msgstring,"New client %s",inet_ntoa(addr.sin_addr));
        MESSAGE(msgstring);
        }
      clnt=InsertClientList(list);
      if (clnt == NULL)
        {
        PERROR("malloc()");
        retval = -1;
        }
      else
	{
        clnt->addr=addr;
        clnt->addrlen=len;
        strcpy(clnt->magic, buf.magic);
	}
      }
    }

  if (retval >= 0)
    {
    /* tell the current status to client either matched or new */
    strcpy(buf.magic, clnt->magic);
    buf.cmd=TranslateAction(current_status, buf.magic);
    count=sendto(fd, &buf, sizeof(buf), 0, (struct sockaddr *)&addr, len);
    if (count < 0)
      {
      PERROR("sendto()");
      retval = -1;
      }
    else if (count != sizeof(buf))
      {
      sprintf(msgstring,"invalid message size (%i) to %s",
  	      count, inet_ntoa(addr.sin_addr));
      MESSAGE(msgstring);
      retval = -1;
      }
    else
      PDEBUG_L9((stderr,"%s: Sent %i\n",CurrTime(),buf.cmd));
    }

  return retval;
}
  
/*-------------------------------------------------------------------*/
/* Alloc memory for struct client at begin of list.
 * Return the pointer of the allocated area or NULL if memory exausted
 */

static struct client *InsertClientList(struct client **p2Client)
{
  struct client *retval=NULL;
  
  retval=malloc(sizeof(struct client));
  if (retval != NULL)
    {
    retval->next=*p2Client;
    *p2Client=retval;
    }
  
  return retval;
}

/*-------------------------------------------------------------------*/
/* Free memory from client list
 */

static void FreeClientList(struct client *pClient)
{
  if (pClient != NULL)
    {
    FreeClientList(pClient->next);
    free(pClient);
    }
}

/*-------------------------------------------------------------------*/
/* Send to clients the new power status.
 * Return negative number on error.
 */

static int TellClients(int action, int serverfd, struct client *clPtr)
{
  int retval=0;
  int i;
  struct powerd_message buf;
  struct client *ptr;

  if (action != ACT_OK && action != ACT_FAIL && action != ACT_FAILLOW)
    retval = -1;
  else
    {
    for (i=0; i < 3; i++)  /* three times: isn't udp unreliable? */
      {
      for(ptr=clPtr; ptr != NULL; ptr=ptr->next)
        {
        buf.cmd=TranslateAction(action, ptr->magic);
        strcpy(buf.magic, ptr->magic);
        sendto(serverfd,&buf, sizeof(buf), 0,
  	       (struct sockaddr *)&(ptr->addr), ptr->addrlen);
        PDEBUG_L9((stderr,"%s: Told %s about\n",CurrTime(),
	        inet_ntoa(ptr->addr.sin_addr)));
        }
      sleep(1);
      }
    }

  return retval;
}

/*-------------------------------------------------------------------*/
/*
 * This function, finally, is in charge of speaking to init. It is
 * system dependent, and systemwide configuration is documented in
 * README.`uname` in the source directory. (Not yet...)
 */

static int PowerFail(int action)
{
  int retval=0;
  int exitnow=FALSE;
  int mounted_fs;
  int fd;
  char cmd[256];		/* command buffer */

  PDEBUG_L9((stderr,"%s: PowerFail(\'%c\')\n",CurrTime(),action));

  if (action==ACT_OK)           MESSAGE("Power is back");
  else if (action==ACT_FAIL)    MESSAGE("Power is failing");
  else if (action==ACT_FAILLOW) MESSAGE("Power is failing and battery is low");
  else
    {
    exitnow=TRUE;
    retval=0;
    }

  if (exitnow != TRUE)
    {
    /* do it now, and do it quick. */
    if (options[O_DISABLED].iv)
      {
      char s[64];
      sprintf(s,"PowerFail(\'%c\') -- disabled",action);
      MESSAGE(s);
      exitnow=TRUE;
      retval=0;
      }
    }
    
#ifdef linux                 /******** This is the linux way to do it *******/
  if (exitnow != TRUE)
    {
    if ((mounted_fs=GetMountFS(options[O_POWERFILE].sv)) == FS_RO)
      {
      /* We need to remount filesystem with r/w option */
      sprintf(cmd, MOUNT_FS_RW, MountPointOf(options[O_POWERFILE].sv));
      MESSAGE(cmd);
      system(cmd);
      }
     
    /* Create an info file for init. */
    unlink(options[O_POWERFILE].sv);
    if ((fd=open(options[O_POWERFILE].sv, O_CREAT|O_WRONLY, 0644)) < 0)
      {
      PERROR(options[O_POWERFILE].sv);
      exitnow=TRUE;
      retval=1; /* maybe next time... */
      }
    else
      {
      if (action==ACT_OK)
        write(fd, "OK\n", sizeof("OK\n")-1);
      else if (action==ACT_FAILLOW)
        write(fd, "LOW\n", sizeof("LOW\n")-1);
      else
        write(fd, "FAIL\n", sizeof("FAIL\n")-1);
      close(fd);
      }

    if (mounted_fs == FS_RO)
      {
      /* We need to remount filesystem with ro option */
      sprintf(cmd, MOUNT_FS_RO, MountPointOf(options[O_POWERFILE].sv));
      MESSAGE(cmd);
      system(cmd);
      }
    }
    

  if (exitnow != TRUE)
    {
    MESSAGE("Telling init about...");
    if (kill(1, SIGPWR) < 0)
      {
      PERROR("kill(1, SIGPWR)");
      exitnow=TRUE;
      retval=1;
      }
    }
#else                                 /******** Open to enhancments *******/
  if (exitnow != TRUE)
    {
  	    f=popen("wall","w");
    if (!f) f=popen("/etc/wall","w");
    if (!f) f=popen("/sbin/wall","w");
    if (!f) f=stderr;
    fprintf(f,"The power is going %s\n", action==ACT_OK ? "up" : 
        (action==ACT_FAIL ? "down" : "down (and battery low)"));
    if (f!=stderr) pclose(f);
    }
#endif
  
  return retval;
}

